/*
 * Riven save file manipulator.
 *
 * David I. Bell
 * dbell@canb.auug.org.au
 * February 15, 1998
 */

#include <stdio.h>


#ifdef	__STDC__
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#define	PROTO(a)	a
#else
#define	PROTO(a)	()
#define	const
#endif


typedef	int	BOOL;

#define	TRUE	((BOOL) 1)
#define	FALSE	((BOOL) 0)


/*
 * Special file offsets used in several places.
 */
#define	ISLAND		((long) 00146)
#define	POSITION	((long) 04752)
#define	DOMECOMB	((long) 00320)


/*
 * Definition of a variable.
 */
typedef	struct
{
	const char *	name;		/* name of value */
	const char *	desc;		/* description */
	long		offset;		/* offset in save file */
	int		size;		/* number of bytes */
} VAR;


/*
 * Table of known variables.
 */
static const VAR	varTable[] =
{
	/*
	 * Global values.
	 */
	{"island-num",	"Island number",		ISLAND,	2},
	{"island-pos",	"Position on island",		POSITION, 2},
	{"dome-comb",	"Combination for domes",	DOMECOMB, 4},
	{"tel-comb",	"Combination for telescope",	00350,	4},
	{"pris-comb",	"Combination for prison",	00334,	4},
	{"bead-power",	"Fire beads powered on",	00512,	2},
	{"bead-engine",	"Fire beads engine position",	04706,	2},
	{"dome-spin",	"Domes are spinning",		00242,	2},
	{"prison-book",	"Prison book available",	00542,	2},
#if 0
	/*
	 * I don't know what these do but they vary some.
	 */
	{"init2",	"Initial condition 2",		00324,	4},
	{"init3",	"Initial condition 3",		00330,	4},
	{"init5",	"Initial condition 5",		00340,	4},
	{"init6",	"Initial condition 6",		00344,	4},
	{"init8",	"Initial condition 8",		00354,	4},
#endif

	/*
	 * Temple Island related values.
	 */
	{"t-j-ride",	"Temple to Jungle ride present",04406,	2},
	{"t-rot-pos",	"Rotating room position",	04262,	2},
	{"t-rot-ld",	"Rotating room left door open",	04326,	2},
	{"t-rot-rd",	"Rotating room right door open", 04312,	2},
	{"t-rot-bd",	"Rotating room back door open",	04342,	2},
	{"t-walk-steam","Big walkway steam supply on",	04106,	2},
	{"t-walk-hand",	"Big walkway handle up",	01026,	2},
	{"t-walk-pos",	"Big walkway position up",	00652,	2},
	{"t-lift",	"Dome lift position up",	04202,	2},
	{"t-plat",	"Dome inside platform extended",04466,	2},
	{"t-bead-steam","Bead walkway steam supply on",	04672,	2},
	{"t-bead-hand",	"Bead walkway handle up",	04232,	2},
	{"t-bead-pos",	"Bead walkway position up",	04246,	2},
	{"t-tele-steam","Telescope steam supply on",	04562,	2},
	{"t-tele-guard","Telescope guard in place",	04532,	2},
	{"t-proj-door",	"Projector room door open",	04372,	2},
	{"t-proj-open",	"Projector chair is opened",	04026,	2},
	{"t-proj-run",	"Projector chair is running",	04642,	2},
	{"t-big-door",	"Big temple door open",		04612,	2},
	{"t-small-door","Small temple door open",	04452,	2},
	{"t-dome-spin",	"Temple Island dome spinning",	04166,	2},

	/*
	 * Boiler Island related values.
	 */
	{"b-fans",	"Fans are running",		01152,	2},
	{"b-manhole",	"Manhole cover opened",		01436,	2},
	{"b-boiler-door","Boiler door opened",		00716,	2},
	{"b-fd-lock",	"Front office door locked",	01202,	2},
	{"b-fd-open",	"Front office door opened",	01312,	2},
	{"b-bd-lock",	"Back office door locked",	00622,	2},
	{"b-bd-open",	"Back office door opened",	01262,	2},
	{"b-look-door-open", "Lookout door open",	01042,	2},
	{"b-trap-open",	"Trap opened state",		01532,	2},
	{"b-trap-raised","Trap raised",			01442,	2},
	{"b-trap-bait",	"Bait placed in trap",		00636,	2},
	{"b-book-page",	"Page number of office book",	01326,	2},
	{"b-drawer",	"Drawer opened in office",	01136,	2},
	{"b-ball",	"Ball rolled in office",	01562,	2},
	{"b-oven",	"Oven opened in office",	01422,	2},
	{"b-vise",	"Vise closed in office",	01466,	2},
	{"b-dome-spin",	"Boiler Island dome spinning",	01122,	2},

	/*
	 * Jungle Island related values.
	 */
	{"j-t-ride",	"Jungle to Temple ride present", 03446,	2},
	{"j-s-ride",	"Jungle to Spike ride present",	03416,	2},
	{"j-sub-door",	"Door into sub is opened",	03242,	2},
	{"j-sub-pos",	"Location of submarine",	03212,	2},
	{"j-sub-face",	"Direction submarine faces",	03226,	2},
	{"j-sub-turn",	"Direction submarine turns",	03256,	2},
	{"j-sub-plat-1", "Platform 1 is extended",	02522,	2},
	{"j-sub-plat-2", "Platform 2 is extended",	02536,	2},
	{"j-sub-plat-3", "Platform 3 is extended",	02552,	2},
	{"j-sub-plat-4", "Platform 4 is extended",	02566,	2},
	{"j-sub-plat-5", "Platform 5 is extended",	02602,	2},
	{"j-trap-shut",	"Trapeze platform floor shut",	02676,	2},
	{"j-trap-door",	"Trapeze control door opened",	03306,	2},
	{"j-house-door", "Teaching house door opened",	03162,	2},
	{"j-beetle",	"Beetle on gate present",	02456,	2},
	{"j-front-gate","Jungle front gate open",	02712,	2},
	{"j-back-gate",	"Jungle back gate open",	03176,	2},
	{"j-idol",	"Idol jaws open",		03432,	2},
	{"j-pris-door",	"Prison main door open",	03102,	2},
	{"j-pris-grate","Prison drain grate open",	02662,	2},
	{"j-pris-wall",	"Prison secret wall open",	03116,	2},
	{"j-ladder",	"Ladder over lagoon lowered",	03006,	2},
	{"j-pris-present", "Prisoner is present",	03066,	2},
	{"j-stone-door", "Stone room door position",	03322,	2},
	{"j-stone-down", "Stones which are down",	02740,	4},
	{"j-stone-order", "Order of stones being down",	02770,	4},
	{"j-stone-flag","Some stone is down",		03512,	2},
	{"j-stone-enabled", "Stone room is enabled",	03132,	2},
	{"j-stone-bulbs", "Count of light bulbs on",	03336,	2},
	{"j-dome-spin",	"Jungle Island dome spinning",	02646,	2},

	/*
	 * Spike Island related values.
	 */
	{"s-j-ride",	"Spike to Jungle ride present",	01656,	2},
	{"s-b-ride",	"Spike to Boiler ride position",02032,	2},
	{"s-blob",	"Activated water blob",		01752,	2},
	{"s-whark",	"Count of whark visits",	02142,	2},
	{"s-left-right","Right button of left viewer",	02172,	2},
	{"s-cage",	"Water cage is raised",		02126,	2},
	{"s-door",	"Door by ride opened",		02112,	2},
	{"s-dome-spin",	"Spike Island dome spinning",	01642,	2},

	/*
	 * Prison Island related values.
	 */
	{"p-dome-spin",	"Prison Island dome spinning",	03636,	2},

	/*
	 * Moriety related values.
	 */
	{"m-state",	"State of events in Moriety",	03746,	2},

	/*
	 * Gehn's hideout related values.
	 */
	{"g-state",	"State of meetings with Gehn",	00462,	2},
	{"g-cage",	"Gehn's hideout cage open",	02346,	2},

	{NULL,		NULL,				0,	0}
};


/*
 * A position.
 */
typedef	struct
{
	const char *	name;		/* name of location */
	const char *	desc;		/* description of location */
	int		island;		/* location island */
	int		pos;		/* location position */
} POS;


/*
 * Island number values.
 */
#define	GEHN		1	/* Gehn's hideout */
#define	PRISON		2	/* Catherine's prison island */
#define	MORIETY		3	/* Catherine's hidden people */
#define	TEMPLE		4	/* main temple island */
#define	BOILER		5	/* boiler (paper making) island */
#define	SPIKE		6	/* spike (map) island */
#define	JUNGLE		7	/* jungle (village) island */


static const POS	posTable[] =
{
	/*
	 * Temple Island positions.
	 */
	{"t-start",	"Starting point",		TEMPLE, 	155},
	{"t-rot-entry",	"Rotating room main entry",	TEMPLE, 	88},
	{"t-rot",	"Inside rotating room",		TEMPLE,		195},
	{"t-plank",	"Gangplank inside big dome",	TEMPLE, 	11},
	{"t-bead-walk",	"Bead walkway handle",		TEMPLE,		433},
	{"t-cave-gate",	"At gate of cave",		TEMPLE,		107},
	{"t-dome",	"Temple dome",			TEMPLE,		374},
	{"t-bead-steam","Bead walkway steam control",	TEMPLE,		355},
	{"t-walk-steam","Big walkway steam control",	TEMPLE,		406},
	{"t-tele",	"Telescope",			TEMPLE,		137},
	{"t-tele-steam","Telescope steam control",	TEMPLE,		191},
	{"t-j-ride",	"Temple to Jungle ride",	TEMPLE,		46},
	{"t-j-ride-car","Inside Temple to Jungle ride",	TEMPLE,		55},
	{"t-proj-chair","Projector chair",		TEMPLE,		15},
	{"t-temple",	"Temple room",			TEMPLE,		63},
	{"t-lift-bot",	"Lift bottom",			TEMPLE,		398},
	{"t-beads",	"Fire bead controls",		TEMPLE, 	221},

	/*
	 * Jungle Island positions.
	 */
	{"j-b-ride",	"Cart ride to Boiler Island",	JUNGLE, 	610},
	{"j-s-ride",	"Jungle to Spike ride",		JUNGLE,		401},
	{"j-s-ride-car","Inside Jungle to Spike ride",	JUNGLE,		406},
	{"j-t-ride",	"Jungle to Temple ride",	JUNGLE,		808},
	{"j-t-ride-car","Inside Jungle to Spike ride",	JUNGLE,		813},
	{"j-beach",	"Beach", 			JUNGLE, 	645},
	{"j-trap-ctl",	"Trapeze control",		JUNGLE, 	148},
	{"j-house",	"Teaching house",		JUNGLE, 	573},
	{"j-idol",	"In front of Idol",		JUNGLE, 	378},
	{"j-idol-lift",	"Middle level of Idol lift",	JUNGLE, 	393},
	{"j-dome",	"Jungle dome",			JUNGLE, 	353},
	{"j-front-gate", "Jungle front gate",		JUNGLE, 	43},
	{"j-back-gate",	"Jungle back gate",		JUNGLE, 	421},
	{"j-junc",	"Junction of village paths",	JUNGLE, 	626},
	{"j-plat-ctl",	"Sub platform control room",	JUNGLE, 	333},
	{"j-pool",	"Small pool of water",		JUNGLE, 	697},
	{"j-sub-launch","Launching platform for sub",	JUNGLE, 	727},
	{"j-sub-entry",	"Initial entry platform of sub",JUNGLE, 	521},
	{"j-sub-junc",	"Four way junction underwater",	JUNGLE,		533},
	{"j-sub-big",	"Big sub platform",		JUNGLE, 	472},
	{"j-trapeze",	"On the trapeze",		JUNGLE, 	277},
	{"j-village",	"In the village",		JUNGLE, 	497},
	{"j-stone",	"Stone room",			JUNGLE,		215},
	{"j-m-book",	"Book to Moriety",		JUNGLE,		226},

	/*
	 * Boiler Island positions.
	 */
	{"b-j-ride",	"Cart ride to Jungle Island",	BOILER, 	494},
	{"b-s-ride",	"Boiler to Spike ride",		BOILER,		243},
	{"b-s-ride-car","Inside Boiler to Spike ride",	BOILER,		250},
	{"b-t-walkway",	"Walkway to Temple Island",	BOILER, 	105},
	{"b-landing",	"Landing point from cart",	BOILER,		3},
	{"b-water-switch", "Water switch in lake",	BOILER, 	159},
	{"b-controls",	"Controls of boiler",		BOILER, 	28},
	{"b-pipe",	"Inside boiler pipe",		BOILER,		341},
	{"b-dome",	"Boiler dome",			BOILER, 	171},
	{"b-lookout",	"Lookout over lake",		BOILER, 	63},
	{"b-frog-trap",	"Frog trap",			BOILER,		372},
	{"b-office",	"Inside Gehn's office",		BOILER, 	180},

	/*
	 * Spike Island positions.
	 */
	{"s-chair",	"Sitting in viewing chair",	SPIKE,		262},
	{"s-j-ride",	"Spike to Jungle ride",		SPIKE,		229},
	{"s-j-ride-car","Inside Spike to Jungle ride",	SPIKE,		235},
	{"s-dome",	"Spike dome",			SPIKE,		190},
	{"s-lookout",	"Lookout over bubbles",		SPIKE,		205},
	{"s-cage",	"Inside water cage",		SPIKE,		13},
	{"s-3d-map",	"At 3D map machine",		SPIKE,		155},
	{"s-lift",	"Inside outside lift",		SPIKE,		139},
	{"s-b-ride",	"Spike to Boiler ride",		SPIKE,		31},
	{"s-b-ride-far", "Spike to Boiler ride far side", SPIKE,	25},
	{"s-b-ride-car","Inside Spike to Boiler ride",	SPIKE,		30},

	/*
	 * Prison Island positions.
	 */
	{"p-dome",	"Prison Island dome",		PRISON,		39},
	{"p-lock",	"Lock for prison",		PRISON,		29},
	{"p-room",	"Catherine's prison room",	PRISON,		32},

	/*
	 * Gehn's hideout positions.
	 */
	{"g-hideout",	"Within Gehn's hideout",	GEHN,		47},
	{"g-cage",	"Cage at Gehn's hideout",	GEHN,		1},
	{"g-toy",	"Toy at Gehn's hideout",	GEHN,		84},

	/*
	 * Moriety hideout positions.
	 */
	{"m-view",	"View of Moriety",		MORIETY,	3},
	{"m-prison",	"Prison on Moriety",		MORIETY,	6},

	{NULL,		NULL,				0,		0}
};


/*
 * Table of commands.
 */
typedef	struct
{
	const char *	name;
	BOOL		needsFileNameFlag;
	BOOL		nestedOkFlag;
	BOOL		(*func) PROTO((int, const char **));
} COMMAND;


static	BOOL	CommandShow PROTO((int, const char **));
static	BOOL	CommandSet PROTO((int, const char **));
static	BOOL	CommandPos PROTO((int, const char **));
static	BOOL	CommandDome PROTO((int, const char **));
static	BOOL	CommandPrompt PROTO((int, const char **));
static	BOOL	CommandListVar PROTO((int, const char **));
static	BOOL	CommandListPos PROTO((int, const char **));
static	BOOL	CommandQuit PROTO((int, const char **));
static	BOOL	CommandHelp PROTO((int, const char **));


static	COMMAND	commandTable[] =
{
	{"show",	TRUE,	TRUE,	CommandShow},
	{"set",		TRUE,	TRUE,	CommandSet},
	{"pos",		TRUE,	TRUE,	CommandPos},
	{"dome",	TRUE,	TRUE,	CommandDome},
	{"prompt",	TRUE,	FALSE,	CommandPrompt},
	{"listvar",	FALSE,	TRUE,	CommandListVar},
	{"listpos",	FALSE,	TRUE,	CommandListPos},
	{"quit",	FALSE,	TRUE,	CommandQuit},
	{"help",	FALSE,	TRUE,	CommandHelp},
	{NULL,		FALSE,	FALSE,	0}
};


/*
 * Local procedures.
 */
static	BOOL		Command PROTO((int, const char ** argv, BOOL));
static	const VAR *	FindVar PROTO((const char *));
static	void		PrintAllVars PROTO((void));
static	BOOL		PrintVar PROTO((const char *));
static	BOOL		WriteVar PROTO((const char *, const char *));
static	void		PrintPosition PROTO((void));
static	BOOL		SetPosition PROTO((const char *));
static	void		PrintDomeCombination PROTO((void));
static	BOOL		SetDomeCombination PROTO((const char *, const char *,
				const char *, const char *, const char *));
static	void		OpenFile PROTO((const char *, int));
static	long		ReadValue PROTO((long, int));
static	void		WriteValue PROTO((long, int, long));
static	BOOL		Contains PROTO((const char *, const char *));
static	long		ParseNumber PROTO((const char * str));


/*
 * File descriptor of Riven save file.
 */
static	int		fd;


/*
 * Filename of Riven save file.
 */
static	const char *	fileName;


/*
 * Flag to indicate quitting.
 */
static	BOOL	quitFlag;


int
main(argc, argv)
	int 		argc;
	const char **	argv;
{
	fd = -1;
	quitFlag = FALSE;
	fileName = NULL;

	argc--;
	argv++;

	if (argc < 1)
	{
		CommandHelp(argc, argv);

		return 1;
	}

	/*
	 * Collect the file name if any.
	 */
	fileName = NULL;

	if ((argc > 0) && (**argv != '-'))
	{
		fileName = *argv++;
		argc--;
	}

	/*
	 * Execute the command.
	 */
	if (!Command(argc, argv, FALSE))
		return 1;

	return 0;
}


/*
 * Execute the specified command which has been broken into its arguments.
 */
static BOOL
Command(argc, argv, nestedFlag)
	int		argc;
	const char **	argv;
	BOOL		nestedFlag;
{
	const char *	cmd;
	const COMMAND *	command;
	BOOL		successFlag;

	/*
	 * Get the command option name if one is there,
	 * stripping the leading dash.
	 */
	cmd = NULL;

	if ((argc > 0) && (**argv == '-'))
	{
		argc--;
		cmd = *argv++;
		cmd++;
	}

	/*
	 * If we are nested and there is no command yet,
	 * then use the first argument as the command.
	 */
	if (nestedFlag && (cmd == NULL) && (argc > 0))
	{
		cmd = *argv++;
		argc--;
	}

	/*
	 * If there is no command then default it to show.
	 */
	if (cmd == NULL)
		cmd = "show";

	/*
	 * Look for the command.
	 */
	for (command = commandTable; command->name; command++)
	{
		if (strcmp(command->name, cmd) == 0)
			break;
	}

	if (command->name == 0)
	{
		fprintf(stderr, "Unknown command \"%s\"\n", cmd);

		return FALSE;
	}

	/*
	 * See if this is a legal command for nesting.
	 */
	if (nestedFlag && !command->nestedOkFlag)
	{
		fprintf(stderr, "Command cannot be nested\n");

		return FALSE;
	}

	/*
	 * See if a file name is required for this command.
	 */
	if ((fileName == NULL) && command->needsFileNameFlag)
	{
		fprintf(stderr, "Filename is required for command\n");

		return FALSE;
	}

	/*
	 * Call the command we found.
	 */
	successFlag = (*command->func)(argc, argv);

	/*
	 * Close the file if it was opened.
	 */
	if (fd >= 0)
	{
		(void) close(fd);

		fd = -1;
	}

	return successFlag;
}


static BOOL
CommandQuit(argc, argv)
	int		argc;
	const char **	argv;
{
	quitFlag = TRUE;

	return TRUE;
}


static BOOL
CommandHelp(argc, argv)
	int		argc;
	const char **	argv;
{
	printf("Usage: riv filename -show [var] ...\n");
	printf("   or: riv filename -set [var value] ...\n");
	printf("   or: riv filename -pos [location]\n");
	printf("   or: riv filename -dome [combination]\n");
	printf("   or: riv filename -prompt\n");
	printf("   or: riv -listvar [word] ...\n");
	printf("   or: riv -listpos [word] ...\n");
	printf("   or: riv -quit\n");
	printf("   or: riv -help\n");
	printf("\n");
	printf("When run using -prompt, commands are given without any\n");
	printf("program name or file name, and the leading dash can be\n");
	printf("omitted.  When run from the command line, the -show\n");
	printf("option can be omitted.\n");

	return TRUE;
}


static BOOL
CommandShow(argc, argv)
	int		argc;
	const char **	argv;
{
	OpenFile(fileName, 0);

	if (argc == 0)
	{
		PrintAllVars();

		return TRUE;
	}

	while (argc-- > 0)
	{
		if (!PrintVar(*argv++))
			return FALSE;
	}

	return TRUE;
}


static BOOL
CommandSet(argc, argv)
	int		argc;
	const char **	argv;
{
	if ((argc & 0x1) != 0)
	{
		fprintf(stderr, "Even number of arguments required for -set\n");

		return FALSE;
	}

	OpenFile(fileName, 2);

	while (argc > 0)
	{
		if (!WriteVar(argv[0], argv[1]))
			return FALSE;

		argc -= 2;
		argv += 2;
	}

	return TRUE;
}


static BOOL
CommandPos(argc, argv)
	int		argc;
	const char **	argv;
{
	if (argc == 0)
	{
		OpenFile(fileName, 0);

		PrintPosition();

		return TRUE;
	}

	if (argc != 1)
	{
		fprintf(stderr, "Wrong number of args for -pos\n");

		return FALSE;
	}

	OpenFile(fileName, 2);

	return SetPosition(*argv);
}


static BOOL
CommandDome(argc, argv)
	int		argc;
	const char **	argv;
{
	if (argc == 0)
	{
		OpenFile(fileName, 0);

		PrintDomeCombination();

		return TRUE;
	}

	if (argc != 5)
	{
		fprintf(stderr, "Wrong number of args for -dome\n");

		return FALSE;
	}

	OpenFile(fileName, 2);

	return SetDomeCombination(argv[0], argv[1], argv[2], argv[3], argv[4]);
}


static BOOL
CommandPrompt(oldArgc, oldArgv)
	int		oldArgc;
	const char **	oldArgv;
{
	char *		cp;
	int		len;
	int		argc;
	const char *	argv[128];
	char		buf[256];

	/*
	 * Test the existence of the file.
	 */
	OpenFile(fileName, 0);

	close(fd);
	fd = -1;

	/*
	 * Keep reading commands until an end of file or quit command.
	 */
	while (!quitFlag)
	{
		printf("riv> ");
		fflush(stdout);

		if (fgets(buf, sizeof(buf), stdin) == NULL)
		{
			printf("\n");

			break;
		}

		len = strlen(buf);

		if ((len > 0) && (buf[len - 1] == '\n'))
			buf[--len] = '\0';

		argc = 0;
		cp = buf;

		/*
		 * Collect arguments.
		 */
		for (;;)
		{
			while ((*cp == ' ') || (*cp == '\t'))
				cp++;

			if (*cp == '\0')
				break;

			argv[argc++] = cp;

			while (*cp && (*cp != ' ') && (*cp != '\t'))
				cp++;

			if (*cp)
				*cp++ = '\0';
		}

		/*
		 * Execute the command if there was one.
		 */
		if (argc > 0)
			Command(argc, argv, TRUE);

		/*
		 * Close the file if it was opened.
		 */
		if (fd >= 0)
		{
			(void) close(fd);
			fd = -1;
		}
	}

	return TRUE;
}


static BOOL
CommandListVar(argc, argv)
	int		argc;
	const char **	argv;
{
	const VAR *	var;
	int		loopArgc;
	const char **	loopArgv;
	BOOL		showFlag;

	/*
	 * If no words are given, show all vars.
	 */
	if (argc <= 0)
	{
		for (var = varTable; var->name; var++)
			printf("%s: %s\n", var->name, var->desc);

		return TRUE;
	}

	/*
	 * Iterate over each var, showing it if it matches one of the
	 * words.
	 */
	for (var = varTable; var->name; var++)
	{
		loopArgc = argc;
		loopArgv = argv;
		showFlag = FALSE;

		while (loopArgc-- > 0)
		{
			if (Contains(var->name, *loopArgv) ||
				Contains(var->desc, *loopArgv))
			{
				showFlag = TRUE;
			}

			loopArgv++;
		}

		if (showFlag)
			printf("%s: %s\n", var->name, var->desc);
	}

	return TRUE;
}


static BOOL
CommandListPos(argc, argv)
	int		argc;
	const char **	argv;
{
	const POS *	pos;
	int		loopArgc;
	const char **	loopArgv;
	BOOL		showFlag;

	/*
	 * If no words are given, show all positions.
	 */
	if (argc <= 0)
	{
		for (pos = posTable; pos->name; pos++)
			printf("%s: %s\n", pos->name, pos->desc);

		return TRUE;
	}

	/*
	 * Iterate over each position, showing it if it matches one of the
	 * words.
	 */
	for (pos = posTable; pos->name; pos++)
	{
		loopArgc = argc;
		loopArgv = argv;
		showFlag = FALSE;

		while (loopArgc-- > 0)
		{
			if (Contains(pos->name, *loopArgv) ||
				Contains(pos->desc, *loopArgv))
			{
				showFlag = TRUE;
			}

			loopArgv++;
		}

		if (showFlag)
			printf("%s: %s\n", pos->name, pos->desc);
	}

	return TRUE;
}


static BOOL
WriteVar(name, strValue)
	const char *	name;
	const char *	strValue;
{
	const VAR *	var;
	const char *	str;
	long		value;

	value = 0;

	str = strValue;

	while ((*str >= '0') && (*str <= '9'))
		value = value * 10 + *str++ - '0';

	if (*str)
	{
		fprintf(stderr, "Illegal number for value\n");

		return FALSE;
	}

	var = FindVar(name);

	if (var == NULL)
	{
		fprintf(stderr, "Var \"%s\" is unknown\n", name);

		return FALSE;
	}

	WriteValue(var->offset, var->size, value);

	return TRUE;
}


static void
PrintAllVars()
{
	const VAR *	var;

	for (var = varTable; var->name; var++)
		(void) PrintVar(var->name);
}


static BOOL
PrintVar(name)
	const char *	name;
{
	const VAR *	var;
	long		value;

	var = FindVar(name);

	if (var == NULL)
	{
		fprintf(stderr, "%s: unknown\n", name);

		return FALSE;
	}

	value = ReadValue(var->offset, var->size);

	printf("%s: %ld\n", name, value);

	return TRUE;
}


static const VAR *
FindVar(name)
	const char *	name;
{
	const VAR *	var;

	for (var = varTable; var->name; var++)
	{
		if (strcmp(name, var->name) == 0)
			return var;
	}

	return NULL;
}


/*
 * Print the current position.
 */
static void
PrintPosition()
{
	const POS *	pos;
	const char *	name;
	int		island;
	int		position;

	island = ReadValue(ISLAND, 2);
	position = ReadValue(POSITION, 2);

	/*
	 * Try to find the position in our table.
	 */
	for (pos = posTable; pos->name; pos++)
	{
		if ((pos->island == island) && (pos->pos == position))
		{
			printf("%s (%d,%d): %s\n", pos->name, island,
				position, pos->desc);

			return;
		}
	}

	/*
	 * It isn't in the table, print it numerically.
	 */
	switch (island)
	{
		case GEHN:	name = "Gehn"; break;
		case PRISON:	name = "Prison"; break;
		case MORIETY:	name = "Moriety"; break;
		case TEMPLE:	name = "Temple"; break;
		case BOILER:	name = "Boiler"; break;
		case SPIKE:	name = "Spike"; break;
		case JUNGLE:	name = "Jungle"; break;
		default:	name = "Unknown"; break;
	}

	printf("%s (%d), position %d\n", name, island, position);
}


/*
 * Set the current position.
 */
static BOOL
SetPosition(position)
	const char *	position;
{
	const POS *	pos;

	for (pos = posTable; pos->name; pos++)
	{
		if (strcmp(pos->name, position) == 0)
		{
			WriteValue(ISLAND, 2, pos->island);
			WriteValue(POSITION, 2, pos->pos);

			return TRUE;
		}
	}

	fprintf(stderr, "Position %s is not known\n", position);

	return FALSE;
}


static void
PrintDomeCombination()
{
	long	comb;
	int	digit;

	comb = ReadValue(DOMECOMB, 4);

	for (digit = 1; digit <= 25; digit++)
	{
		if ((comb >> (25 - digit)) & 0x01)
			printf("%3d", digit);
	}

	printf("\n");
}


static BOOL
SetDomeCombination(str1, str2, str3, str4, str5)
	const char *	str1;
	const char *	str2;
	const char *	str3;
	const char *	str4;
	const char *	str5;
{
	long	comb;
	int	i;
	int	digits[6];

	comb = 0;

	digits[0] = ParseNumber(str1);
	digits[1] = ParseNumber(str2);
	digits[2] = ParseNumber(str3);
	digits[3] = ParseNumber(str4);
	digits[4] = ParseNumber(str5);
	digits[5] = 100;

	for (i = 0; i < 5; i++)
	{
		if ((digits[i] < 1) || (digits[i] > 25) ||
			(digits[i] >= digits[i + 1]))
		{
			fprintf(stderr, "Bad combination\n");

			return FALSE;
		}

		comb |= 0x01 << (25 - digits[i]);
	}

	WriteValue(DOMECOMB, 4, comb);

	return TRUE;
}


/*
 * Return whether the haystack string contains the needle string
 * somewhere within it, while ignoring upper case/lower case differences.
 */
static BOOL
Contains(hayStack, needle)
	const char *	hayStack;
	const char *	needle;
{
	const char *	cp1;
	const char *	cp2;
	int		ch1;
	int		ch2;
	int		hayStackLength;
	int		needleLength;
	int		len;

	hayStackLength = strlen(hayStack);
	needleLength = strlen(needle);

	while (hayStackLength >= needleLength)
	{
		cp1 = hayStack;
		cp2 = needle;
		len = needleLength;

		while (len > 0)
		{
			ch1 = *cp1++;
			ch2 = *cp2++;

			if ((ch1 >= 'A') && (ch1 <= 'Z'))
				ch1 += 'a' - 'A';

			if ((ch2 >= 'A') && (ch2 <= 'Z'))
				ch2 += 'a' - 'A';

			if (ch1 != ch2)
				break;

			len--;
		}

		if (len == 0)
			return TRUE;

		hayStack++;
		hayStackLength--;
	}

	return FALSE;
}


static void
OpenFile(name, mode)
	const char *	name;
	int		mode;
{
	if (name == NULL)
	{
		fprintf(stderr, "No filename specified\n");

		exit(1);
	}

	fd = open(name, mode);

	if (fd < 0)
	{
		fprintf(stderr, "Cannot open \"%s\" for %s\n", name,
			(mode == 0) ? "reading" : "writing");

		exit(1);
	}
}


static long
ReadValue(offset, size)
	long	offset;
	int	size;
{
	int		cc;
	int		i;
	long		value;
	unsigned char	data[4];

	if ((size != 1) && (size != 2) && (size != 4))
	{
		fprintf(stderr, "Unsupported value size %d\n", size);

		exit(1);
	}

	if (lseek(fd, offset, 0) < 0)
	{
		fprintf(stderr, "lseek failed\n");

		exit(1);
	}

	cc = read(fd, data, size);

	if (cc < 0)
	{
		fprintf(stderr, "read failed\n");

		exit(1);
	}

	if (cc < size)
	{
		fprintf(stderr, "Short read\n");

		exit(1);
	}

	value = 0;

	for (i = 0; i < size; i++)
		value = (value << 8) | data[i];

	return value;
}


static void
WriteValue(offset, size, value)
	long	offset;
	int	size;
	long	value;
{
	int	cc;
	int	i;
	char	data[4];

	if ((size != 1) && (size != 2) && (size != 4))
	{
		fprintf(stderr, "Unsupported value size %d\n", size);

		exit(1);
	}

	if (lseek(fd, offset, 0) < 0)
	{
		fprintf(stderr, "lseek failed\n");

		exit(1);
	}

	for (i = size - 1; i >= 0; i--)
	{
		data[i] = value;
		value >>= 8;
	}

	cc = write(fd, data, size);

	if (cc < 0)
	{
		fprintf(stderr, "write failed\n");

		exit(1);
	}

	if (cc < size)
	{
		fprintf(stderr, "Short write\n");

		exit(1);
	}
}


static long
ParseNumber(str)
	const char *	str;
{
	long	value;

	value = 0;

	while ((*str >= '0') && (*str <= '9'))
		value = value * 10 + *str++ - '0';

	if (*str)
		return -1;

	return value;
}
